% current-reinforced random walk for tower of Hanoi maze
% written by Qi Ma  (online Oct 2012)

clear all;
close all;
tic;

myGraph = read_graph('hanoi_exp0.net'); %read in the tower of Hanoi graph(it should be in pajek format)

numv = length(myGraph.nodes.x);   %number of vertices
nume = length(myGraph.edges.source);  %number of edges

out = myGraph.edges.source;
in = myGraph.edges.dest;
lgth = sqrt((myGraph.nodes.x(out)-myGraph.nodes.x(in)).^2 + (myGraph.nodes.y(out)-myGraph.nodes.y(in)).^2);
myGraph.edges.length = lgth';

%% set up the source and sink
[sourcepos,source] = min(myGraph.nodes.x);
[sinkpos,sink] = max(myGraph.nodes.x);
myGraph.nodes.z(source) = 1; 
myGraph.nodes.z(sink) = 0;  

%% output file name and path
subname = sprintf('Hanoi_crrw');
cdir = fullfile(pwd,subname);
if (~exist(cdir,'dir'))
    mkdir(cdir);
end

%% set parameters
Dt = 1e-2;  %time interval for each update
T = 2000;   %total running steps
minsigma = 1e-3;  %minimum conductance (as D_ij in the paper) on each edge
q = 1e-2;  %reinforcement intensity (same q as in the paper)
lambda = 1;  %decay rate (same lambda as in the paper)
bs = 1000;    %externl flow rate (as b in the paper)
extflow = bs*Dt; 


%% initial condition
myGraph.nodes.n = zeros(numv,1); %no particles at the beginning of the simulation
myGraph.edges.sigma = minsigma*ones(nume,1);  %initial conductance of each edge

h=0;
negn_count=0;

%% set up the format of the output figure
currFig = figure;
set(currFig,'Units','centimeters','OuterPosition',[0.5 3 35 19]);
set(currFig,'Position',[0.6   0.8   32  15]);
set(gca,'box','on');  %set the look of the figure
set(gca,'XTick',[]);
set(gca,'YTick',[]);
set(currFig,'PaperUnits','inches');
set(currFig, 'PaperSize',[10 7]);
set(currFig, 'PaperPositionMode', 'auto');


%% main loop
for t=(0:Dt:T)
    real_extflow = poissrnd(extflow);
    myGraph.nodes.n(source) = myGraph.nodes.n(source) +  real_extflow ; %update the number of particles on the source
    myGraph.nodes.n(sink) = max(myGraph.nodes.n(sink) -  real_extflow, 0);  %update the number of particles on the sink
    
    ndiff = myGraph.nodes.n(out) - myGraph.nodes.n(in); %calculate particle difference
    phi = poissrnd(abs(ndiff).*(myGraph.edges.sigma)./(myGraph.edges.length)*(Dt)).* sign(ndiff); %calculate the flow on each edge
    num_p = myGraph.nodes.n;
    
    %% update the number of particles on nodes according to the flow
    des_phi =[];
    [des_phi,edge_pool] = sort(abs(phi),'descend');
    num_change_edge = find(des_phi>0,1,'last');  %find out all the edges whose flow is not zero
    edge_pool = edge_pool(randperm(num_change_edge)); %randomize the order of the edges to update the particles 
    
    for i=1:num_change_edge
        j = edge_pool(i);       
        num_p(out(j))=num_p(out(j))-phi(j);
        num_p(in(j))=num_p(in(j))+phi(j);
        if (num_p(out(j))<0 || num_p(in(j))<0) %if the number of particles become negative, then reduce the flow on the edge so the number of particles on the 'flow-out' node is zero
            negn_count = negn_count +1;
            phi(j) =  sign(phi(j))*(myGraph.nodes.n(out(j))*(num_p(out(j))<0)+myGraph.nodes.n(in(j))*(num_p(in(j))<0));
            myGraph.nodes.n(out(j)) = myGraph.nodes.n(out(j)) - phi(j);
            myGraph.nodes.n(in(j)) = myGraph.nodes.n(in(j)) + phi(j);
            num_p(out(j)) = myGraph.nodes.n(out(j));
            num_p(in(j)) = myGraph.nodes.n(in(j));
        else
            myGraph.nodes.n(out(j)) = num_p(out(j));
            myGraph.nodes.n(in(j)) = num_p(in(j));
        end
    end
            
    %% update conductivity (D_ij) for each edge
    myGraph.edges.sigma = myGraph.edges.sigma + q*abs(phi)- lambda*myGraph.edges.sigma*Dt;
    myGraph.edges.sigma = max(myGraph.edges.sigma,minsigma);
    
    if (mod(t,50)==0) %save the graph each 10 unit of time
        t
        h=h+1;
        GraphStore(h) = myGraph;
        GraphStore_t(h) = t;
    end

end
filename = sprintf('HanoiData_crrw.mat');
save(filename,'GraphStore');
movefile(filename, cdir);

%% output all the figures
for i=1:h
    figname = sprintf('hanoi_crrw_%04d.png',GraphStore_t(i),GraphStore_t(i));
    plot_sigma_with_n(GraphStore(i), currFig);
    print(gcf,'-dpng','-r100',figname);
    movefile(figname,cdir);
end

toc;
